 /*
  * mod_auth_ldap
  *
  * User Authentication against LDAP database
  *
  * Alexander Mayrhofer axelm+ldap@nona.net
  *
  * http://nona.net/software/ldap/
  *
  */

 /* 
  * MODULE-DEFINITION-START
  * Name: ldap_auth_module
  * ConfigStart
    LDAP_LIBS="-lldap -llber"
    LIBS="$LIBS $LDAP_LIBS"
    echo "       + using LDAP libraries for auth_ldap module"
  * ConfigEnd
  * MODULE-DEFINITION-END
  */

#include "httpd.h"
#include "http_config.h"
#include "http_core.h"
#include "http_log.h"
#include "http_protocol.h"
#include <lber.h>
#include <ldap.h>
#include <unistd.h>

#define SHA1_DIGEST_LENGTH 29 /* sha1 digest length including null character */

typedef struct  {

    char *auth_ldaphosts;
    char *auth_ldapbinddn;
    char *auth_ldapbindpassword;
    char *auth_ldapbasedn;
    char *auth_ldapuserkey;
    char *auth_ldappasskey;
    char *auth_ldapgroupkey;
    char *auth_ldapsearchscope;
    int   auth_ldapauthoritative;
    int   auth_ldapbasesearch;
    int   auth_ldapcryptpasswords;
    int   auth_ldapbindasuser;
    int   auth_ldapschemeprefix;

} ldap_auth_config_rec;

void *create_ldap_auth_dir_config (pool *p, char *d)
{
    ldap_auth_config_rec *sec
       = (ldap_auth_config_rec *)ap_pcalloc (p, sizeof(ldap_auth_config_rec));

    sec->auth_ldaphosts = NULL;
    sec->auth_ldapbinddn = NULL;
    sec->auth_ldapbindpassword = NULL;
    sec->auth_ldapbasedn = NULL;
    sec->auth_ldapsearchscope = NULL;
    sec->auth_ldapauthoritative = 1; /* fortress is secure by default */
    sec->auth_ldapbasesearch = 0; 
    sec->auth_ldapcryptpasswords = 0; 
    sec->auth_ldapbindasuser = 0; /* by default, we bind as configured user to ldap */
    sec->auth_ldapschemeprefix = 0; /* no scheme prefix in password strings */

    return sec;
}

command_rec ldap_auth_cmds[] = {
{ "AuthLDAPHosts", ap_set_string_slot,
    (void*)XtOffsetOf(ldap_auth_config_rec, auth_ldaphosts),
    OR_AUTHCFG, TAKE1, 
    "List of LDAP hosts which should be queried" },
{ "AuthLDAPBindDN", ap_set_string_slot,
    (void*)XtOffsetOf(ldap_auth_config_rec, auth_ldapbinddn),
    OR_AUTHCFG, TAKE1, 
    "As which DN to bind to the LDAP directory" },
{ "AuthLDAPBindPassword", ap_set_string_slot,
    (void*)XtOffsetOf(ldap_auth_config_rec, auth_ldapbindpassword),
    OR_AUTHCFG, TAKE1, 
    "The password corresponding to AuthLDAPBindDN" },
{ "AuthLDAPBaseDN", ap_set_string_slot,
    (void*)XtOffsetOf(ldap_auth_config_rec, auth_ldapbasedn),
    OR_AUTHCFG, TAKE1, 
    "The DN in the LDAP directory which contains the per-user subnodes" },
{ "AuthLDAPUserKey", ap_set_string_slot,
    (void*)XtOffsetOf(ldap_auth_config_rec, auth_ldapuserkey),
    OR_AUTHCFG, TAKE1, 
    "The key which's value contains the username in the directory" },
{ "AuthLDAPPassKey", ap_set_string_slot,
    (void*)XtOffsetOf(ldap_auth_config_rec, auth_ldappasskey),
    OR_AUTHCFG, TAKE1, 
    "The key to the user's password" },
{ "AuthLDAPGroupKey", ap_set_string_slot,
    (void*)XtOffsetOf(ldap_auth_config_rec, auth_ldapgroupkey),
    OR_AUTHCFG, TAKE1, 
    "the key to the user's group membership string" },
{ "AuthLDAPBaseSearch", ap_set_flag_slot,
    (void*)XtOffsetOf(ldap_auth_config_rec, auth_ldapbasesearch),
    OR_AUTHCFG, FLAG, 
    "Deprecated. Use 'AuthLDAPSearchScope'. Anyway: Set to 'on' if you do want to do a base search only" },
{ "AuthLDAPSearchScope", ap_set_string_slot,
    (void*)XtOffsetOf(ldap_auth_config_rec, auth_ldapsearchscope),
    OR_AUTHCFG, TAKE1, 
    "Set Scope when searching in LDAP. Can be 'base', 'onelevel' or 'subtree'" },
{ "AuthLDAPAuthoritative", ap_set_flag_slot,
    (void*)XtOffsetOf(ldap_auth_config_rec, auth_ldapauthoritative),
    OR_AUTHCFG, FLAG, 
    "Set to 'off' to allow control to be passed on, if UserID unknow to this module" },
{ "AuthLDAPCryptPasswords", ap_set_flag_slot,
    (void*)XtOffsetOf(ldap_auth_config_rec, auth_ldapcryptpasswords),
    OR_AUTHCFG, FLAG, 
    "Set to 'on' if you do have crypted password strings" },
{ "AuthLDAPBindAsUser", ap_set_flag_slot,
    (void*)XtOffsetOf(ldap_auth_config_rec, auth_ldapbindasuser),
    OR_AUTHCFG, FLAG, 
    "Set to 'on' if you want to bind as the user for authentication" },
{ "AuthLDAPSchemePrefix", ap_set_flag_slot,
    (void*)XtOffsetOf(ldap_auth_config_rec, auth_ldapschemeprefix),
    OR_AUTHCFG, FLAG, 
    "Set to 'on' if you have scheme-prefixed password strings as described in rfc2307" },
{ NULL }
};

module ldap_auth_module;

char *get_ldap_val(request_rec *r, char *user, char *pass, ldap_auth_config_rec *conf, char *attr) {
    char *bindas, *bindpass;
    LDAP *ldap;
    LDAPMessage *ld_result, *ld_entry;
    char ldap_query[200];
    char ldap_base[200];
    char **ld_values;
    char *retval = NULL;
    int  ldap_scope;

    if(!(ldap = ldap_open(conf->auth_ldaphosts, LDAP_PORT))) { 
        printf("ldap host: %s", conf->auth_ldaphosts);
        ap_log_reason("ldap_open failed", conf->auth_ldaphosts, r);
        return NULL;
        }
    if(conf->auth_ldapbindasuser) {
       if ((!pass) || (!strcmp(pass,""))) {
          ap_log_reason("ldap_bind: no password given (BindAsUser enabled)!", NULL, r);
          return NULL;
          }
       ap_snprintf(ldap_base, sizeof ldap_base, "%s=%s,%s", conf->auth_ldapuserkey, user, conf->auth_ldapbasedn);
       bindas = ldap_base;
       bindpass = pass;
       }
    else {
       bindas = conf->auth_ldapbinddn;
       bindpass = conf->auth_ldapbindpassword;
       }
    if(ldap_simple_bind_s(ldap, bindas, bindpass)) {
        ap_log_reason("ldap_bind failed", bindas, r);
        return NULL;
        }
    if(!attr) {
	ldap_unbind_s(ldap);
	return "bind";
	}
    ap_snprintf(ldap_query, sizeof ldap_query, "%s=%s", conf->auth_ldapuserkey, user);
    /* check for search scope */
    if(conf->auth_ldapbasesearch || (conf->auth_ldapsearchscope && 
                  !strcasecmp(conf->auth_ldapsearchscope, "base"))) {
       ap_snprintf(ldap_base, sizeof ldap_base, "%s=%s,%s", conf->auth_ldapuserkey, user, conf->auth_ldapbasedn);
       ldap_scope = LDAP_SCOPE_BASE;
       }
    else {
       ap_snprintf(ldap_base, sizeof ldap_base, "%s", conf->auth_ldapbasedn);
       if(conf->auth_ldapsearchscope) {
           ldap_scope = strcasecmp(conf->auth_ldapsearchscope, "subtree") ?
                        LDAP_SCOPE_ONELEVEL : LDAP_SCOPE_SUBTREE; 
           }
       }
    if(ldap_search_s(ldap, ldap_base, ldap_scope,
                     ldap_query, NULL, 0, &ld_result)) {
        ap_log_reason("ldap_search failed", ldap_query, r);
        ldap_unbind_s(ldap);
        return NULL;
        }
    if(!(ld_entry = ldap_first_entry(ldap, ld_result))) {
        ap_log_reason("user not found", user, r);
        ldap_unbind_s(ldap);
        return NULL;
        }
    if(!(ld_values = ldap_get_values(ldap, ld_entry, attr))) {
        ap_log_reason("ldap_get_values failed", attr, r);
        ldap_unbind_s(ldap);
        return NULL;
        }
    
    retval = ap_palloc (r->pool, strlen(ld_values[0])+1);
    strcpy(retval,ld_values[0]);

    ldap_value_free(ld_values);
    ldap_unbind_s(ldap);

    return retval; 
}

char  *get_ldap_grp(request_rec *r, char *user, char *pass, ldap_auth_config_rec *conf) {
    return get_ldap_val(r, user, pass, conf, conf->auth_ldapgroupkey);
}

int ldap_authenticate_basic_user (request_rec *r)
{
    ldap_auth_config_rec *sec =
      (ldap_auth_config_rec *)ap_get_module_config (r->per_dir_config,
						&ldap_auth_module);
    conn_rec *c = r->connection;
    char *sent_pw, *real_pw, *digest;
    char errstr[MAX_STRING_LEN];
    int res;

    if ((res = ap_get_basic_auth_pw (r, (const char **)&sent_pw)))
        return res;

    if(!sec->auth_ldapuserkey)
        return DECLINED;
    
    if(sec->auth_ldapbindasuser) {
        if(!(real_pw = get_ldap_val(r, c->user, sent_pw, sec, sec->auth_ldappasskey))) {
	    if (!(sec->auth_ldapauthoritative))
	        return DECLINED;
            ap_snprintf(errstr, sizeof(errstr), "LDAP user %s not found or password invalid", c->user);
	    ap_log_reason (errstr, r->filename, r);
            ap_note_basic_auth_failure (r);
            return AUTH_REQUIRED;
            }
        return OK;
        }
    if(!(real_pw = get_ldap_val(r, c->user, sent_pw, sec, sec->auth_ldappasskey))) {
	if (!(sec->auth_ldapauthoritative))
	    return DECLINED;
        ap_snprintf(errstr, sizeof(errstr), "LDAP user %s not found", c->user);
	ap_log_reason (errstr, r->filename, r);
	ap_note_basic_auth_failure (r);
	return AUTH_REQUIRED;
    }    
    if (!sec->auth_ldapbindasuser) { 
        /* We didn't use the credentials to bind to LDAP, so check them.. */
        if (sec->auth_ldapschemeprefix) {
           /* we do use scheme prefixes as described in RFC2307 */
           /* currently sha, crypt are supported */
           if (!strncasecmp("{crypt}", real_pw, 7)) {
              if(!strcmp(real_pw+7, crypt(sent_pw,real_pw+7))) 
		 return OK;
              }
           if (!strncasecmp("{sha}", real_pw, 5)) {
              digest = ap_palloc (r->pool, SHA1_DIGEST_LENGTH);
              sha1_digest(sent_pw, strlen(sent_pw), digest);
              if(!strcmp(real_pw+5, digest))
		 return OK;
              }
           ap_snprintf(errstr, sizeof(errstr), 
   	 	    "user %s: password mismatch %s, %s", c->user, real_pw+5, digest);
           ap_log_reason (errstr, r->uri, r);
           ap_note_basic_auth_failure (r);
           return AUTH_REQUIRED;
           }

        if (sec->auth_ldapcryptpasswords) {
           /* Passwords are crypted */
           if(strcmp(real_pw,crypt(sent_pw,real_pw))) {  
              ap_snprintf(errstr, sizeof(errstr), 
	  	    "user %s: password mismatch",c->user);
    	      ap_log_reason (errstr, r->uri, r);
	      ap_note_basic_auth_failure (r);
	      return AUTH_REQUIRED;
              }
           }
        else {
           /* We have clear text passwords ... */
           if(strcmp(real_pw,sent_pw)) {  
               ap_snprintf(errstr, sizeof(errstr), 
	   	    "user %s: password mismatch",c->user);
	       ap_log_reason (errstr, r->uri, r);
	       ap_note_basic_auth_failure (r);
	       return AUTH_REQUIRED;
               }
           }
        }
    return OK;
}
    
/* Checking ID */
    
int ldap_check_auth(request_rec *r) {
    ldap_auth_config_rec *sec =
      (ldap_auth_config_rec *)ap_get_module_config (r->per_dir_config,
						&ldap_auth_module);
    char *user = r->connection->user;
    int m = r->method_number;
    char errstr[MAX_STRING_LEN];
    
    const array_header *reqs_arr = ap_requires (r);
    require_line *reqs = reqs_arr ? (require_line *)reqs_arr->elts : NULL;

    register int x;
    const char *t;
    char *w;

    if (!sec->auth_ldapgroupkey) return DECLINED; 
    if (!reqs_arr) return DECLINED;
    for(x=0; x < reqs_arr->nelts; x++) {
      
	if (! (reqs[x].method_mask & (1 << m))) continue;
	
        t = reqs[x].requirement;
        w = ap_getword(r->pool, &t, ' ');
	
        if(!strcmp(w,"group") && sec->auth_ldapgroupkey) {
           const char *orig_groups,*groups;
	   char *v;

           if (!(groups = get_ldap_grp(r, user, 0, sec))) {
	       if (!(sec->auth_ldapauthoritative))
	           return DECLINED;
               ap_snprintf(errstr, sizeof(errstr), 
			"groups for user %s not found in LDAP directory (key %s)",
			user, sec->auth_ldapgroupkey);
	       ap_log_reason (errstr, r->filename, r);
	       ap_note_basic_auth_failure (r);
	       return AUTH_REQUIRED;
           }
           orig_groups = groups;
           while(t[0]) {
               w = ap_getword(r->pool, &t, ' ');
               groups = orig_groups;
               while(groups[0]) {
                   v = ap_getword(r->pool, &groups,',');
                   if(!strcmp(v,w))
                       return OK;
               }
           }
           ap_snprintf(errstr, sizeof(errstr), 
		"user %s not in right group",user);
	   ap_log_reason (errstr, r->filename, r);
           ap_note_basic_auth_failure(r);
	   return AUTH_REQUIRED;
       }
    }
    
    return DECLINED;
}


module ldap_auth_module = {
   STANDARD_MODULE_STUFF,
   NULL,			/* initializer */
   create_ldap_auth_dir_config,	/* dir config creater */
   NULL,			/* dir merger --- default is to override */
   NULL,			/* server config */
   NULL,			/* merge server config */
   ldap_auth_cmds,		/* command table */
   NULL,			/* handlers */
   NULL,			/* filename translation */
   ldap_authenticate_basic_user,	/* check_user_id */
   ldap_check_auth,		/* check auth */
   NULL,			/* check access */
   NULL,			/* type_checker */
   NULL,			/* fixups */
   NULL,			/* logger */
   NULL				/* header parser */
};
